import logging
from typing import List, Union

import shapely

# noinspection PyUnresolvedReferences
from qgis.core import QgsFeature, QgsGeometry, QgsPointXY

from jord.shapely_utilities import is_multi

logger = logging.getLogger(__name__)


def already_exist_question_mark(iface):
    from shapely.geometry import shape

    layer = iface.activeLayer()
    feature = layer.selectedFeatures()[0]
    shape(feature.__geo_interface__["geometry"])
    shape(feature.geometry().__geo_interface__)


def monkey_patch():
    def mapping_feature(feature):
        geom = feature.geometry()

        properties = {}
        fields = [field.name() for field in feature.fields()]
        properties = dict(zip(fields, feature.attributes()))
        return {
            "type": "Feature",
            "properties": properties,
            "geometry": geom.__geo_interface__,
        }

    def mapping_geometry(geometry):
        geo = geometry.exportToGeoJSON()

        # We have to use eval because exportToGeoJSON() gives us
        # back a string that looks like a dictionary.
        return eval(geo)

    QgsFeature.__geo_interface__ = property(lambda self: mapping_feature(self))
    QgsGeometry.__geo_interface__ = property(lambda self: mapping_geometry(self))


def q_point_creation(
    shapely_geometry: shapely.geometry.base.BaseGeometry,
) -> Union[List[QgsPointXY], List[List[QgsPointXY]]]:
    """
    Construct a List of List or List of QgsPointXY objects.

    :return: Union[List[QgsPointXY], List[List[QgsPointXY]]]
    """

    assert not is_multi(shapely_geometry)

    points = []
    if isinstance(shapely_geometry, shapely.Polygon):
        exterior = []
        for coords in shapely_geometry.exterior.coords:
            points.append(QgsPointXY(*coords))
        points.append(exterior)

        for interior in shapely_geometry.interiors:
            interior_coords = []
            for coords in interior.coords:
                interior_coords.append(QgsPointXY(*coords))
            points.append(interior_coords)

    else:
        for coords in shapely_geometry.coords:
            points.append(QgsPointXY(*coords))

    return points


def q_point_geometry_creation(
    shapely_geometry: shapely.geometry.base.BaseGeometry,
) -> QgsGeometry:
    """
      QgsGeometry object from a basis of QgsPointXY objects.


    fromMultiPointXY	Creates a new geometry from a QgsMultiPointXY object
    fromMultiPolygonXY	Creates a new geometry from a QgsMultiPolygon
    fromMultiPolylineXY	Creates a new geometry from a QgsMultiPolylineXY object
    fromPointXY	Creates a new geometry from a QgsPointXY object
    fromPolygonXY	Creates a new geometry from a QgsPolygon
    fromPolylineXY	Creates a new LineString geometry from a list of QgsPointXY points.

      :return: QgsFeature
    """

    geom = None

    if is_multi(shapely_geometry):
        point_geoms = []
        for geom in shapely_geometry.geoms:
            point_geoms.append(q_point_creation(geom))

        if isinstance(shapely_geometry, shapely.MultiPoint):
            geom = QgsGeometry.fromMultiPointXY(point_geoms)
        elif isinstance(shapely_geometry, shapely.MultiLineString):
            geom = QgsGeometry.fromMultiPolylineXY(point_geoms)
        elif isinstance(shapely_geometry, shapely.MultiPolygon):
            geom = QgsGeometry.fromMultiPolygonXY(point_geoms)
        else:
            logger.error(f"Geometry type {type(shapely_geometry)} is not supported")

    else:
        point_geom = q_point_creation(shapely_geometry)

        if isinstance(shapely_geometry, shapely.Point):
            geom = QgsGeometry.fromPointXY(point_geom)
        elif isinstance(shapely_geometry, shapely.LineString):
            geom = QgsGeometry.fromPolylineXY(point_geom)
        elif isinstance(shapely_geometry, shapely.Polygon):
            geom = QgsGeometry.fromPolygonXY(point_geom)
        else:
            logger.error(f"Geometry type {type(shapely_geometry)} is not supported")

    return geom


def q_point_feature_creation(
    shapely_geometry: shapely.geometry.base.BaseGeometry,
) -> QgsFeature:
    """
    A QgsFeature object from a QgsGeometry object from a basis of QgsPointXY objects.


    :return: QgsFeature
    """

    feat = QgsFeature()
    feat.setGeometry(q_point_geometry_creation(shapely_geometry))
    return feat



from typing import Optional, Sequence, Tuple, Union

from shapely import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)

from jord.shapely_utilities import (
    closest_object,
    explode_polygons,
    extend_line,
    remove_redundant_nodes,
    shift_point,
    split_line,
)

EPSILON = 1e-6

__all__ = [
    "snap_lines",
    "split_lines",
    "intersection_points",
    "clip_lines_with_polygon",
]


def snap_lines(
    lines: Sequence[LineString],
    max_dist: float,
    tolerance: float = 1e-3,
    return_index: bool = False,
) -> Union[Sequence[LineString], Tuple[Sequence[LineString], Sequence[int]]]:
    """
    Snap lines together if the endpoint of one line1 is at most max_dist apart from the line2.
    Both lines are snapped by extending line1 towards line2
    the max distance is measured in the direction of the line1

    :param lines: a list of LineStrings or a MultiLineString
    :param max_dist: maximum distance two endpoints may be joined together
    :param tolerance:
    :param return_index:
    :return:
    """
    import rtree

    # extend all lines with max_dist to snap lines at a sharp angle
    lines_ext = [extend_line(line, max_dist) for line in lines]

    # build spatial index of line bounding boxes
    tree_idx = rtree.index.Index()
    lines_bbox = [l.bounds for l in lines_ext]
    for i, bbox in enumerate(lines_bbox):
        tree_idx.insert(i, bbox)

    lines_snap = []
    idx_snapped = []
    for i1, line in enumerate(lines):
        if isinstance(line, LineString):
            for side in ["start", "end"]:
                coords = line.coords[:]
                # make line extensions
                if side == "start":
                    ext = LineString(
                        [shift_point(coords[0], coords[1], -1.0 * max_dist), coords[0]]
                    )
                    pnt_from = Point(coords[0])
                elif side == "end":
                    ext = LineString(
                        [
                            coords[-1],
                            shift_point(coords[-1], coords[-2], -1.0 * max_dist),
                        ]
                    )
                    pnt_from = Point(coords[-1])

                # point instead of line if max_dist is zero (only clipping)
                if max_dist == 0:
                    ext = pnt_from

                # find close-by lines based on bounds with spatial index
                if tolerance > 0:
                    hits = list(tree_idx.intersection(ext.buffer(tolerance).bounds))
                else:
                    hits = list(tree_idx.intersection(ext.bounds))
                lines_hit = MultiLineString([lines_ext[i] for i in hits if i != i1])
                # find intersection points. function yields list of points
                int_points = intersection_points([ext], lines_hit, tolerance=tolerance)
                if len(int_points) == 0:
                    continue

                # if intersection yields something else then a Point, break down to nearest point
                if len(int_points) > 1:
                    # if more intersections, find closest
                    pnt_to = closest_object(int_points, pnt_from)[0]
                else:
                    pnt_to = int_points[0]

                # at this point pnt_to is the closest intersecting point
                if pnt_from != pnt_to:  # check if lines are not already touching
                    # snap line towards pnt_to
                    if side == "start":
                        coords[0] = pnt_to.coords[0]
                    elif side == "end":
                        coords[-1] = pnt_to.coords[0]
                    line = LineString(coords)
                    if i1 not in idx_snapped:
                        # bookkeeping: list with changes features
                        idx_snapped.append(i1)

            lines_snap.append(line)

    if return_index:
        return lines_snap, idx_snapped

    return lines_snap


def split_lines(
    lines: Sequence[LineString],
    points: Optional[Sequence[Point]] = None,
    tolerance: float = 0.0,
    return_index: bool = False,
) -> Union[Sequence[LineString], Tuple[Sequence[LineString], Sequence[int]]]:
    """
    split lines at intersection, or if given, at points

    :param lines:
    :param points:
    :param tolerance:
    :param return_index:
    :return:
    """
    import rtree

    if isinstance(points, Point):
        points = [points]

    # create output list
    lines_out = []
    index_out = []

    # build spatial index
    if points is None:
        # build spatial index of line bounding boxes
        tree_idx = rtree.index.Index()
        lines_bbox = [l.bounds for l in lines]
        for i, bbox in enumerate(lines_bbox):
            tree_idx.insert(i, bbox)
    else:
        # build spatial index of points
        tree_idx = rtree.index.Index()
        for i, p in enumerate(points):
            tree_idx.insert(i, p.buffer(tolerance).bounds)

    # loop through lines and split lines with split point
    for idx, line in enumerate(lines):
        if points is None:
            # find close-by lines based on bounds with spatial index
            hits = list(tree_idx.intersection(lines_bbox[idx]))
            lines_hit = MultiLineString([lines[i] for i in hits if i != idx])
            # find line intersections
            # lines_other = [l for i, l in enumerate(lines) if i != idx]
            split_points = intersection_points([line], lines_hit, tolerance=tolerance)
        else:
            # find close-by points based on bounds of line with spatial index
            hits = list(tree_idx.intersection(line.bounds))
            points_hit = [points[i] for i in hits]
            # find points which intersects with line
            split_points = [p for p in points_hit if line.distance(p) <= tolerance]
            # split_points = [p for p in points if line.distance(p) <= tolerance]

        # check if intersections for line
        if len(split_points) >= 1:
            for p in split_points:
                line = split_line(line, p, tolerance=tolerance)
            for l in line:
                lines_out.append(l)
                index_out.append(idx)
        else:
            lines_out.append(line)
            index_out.append(idx)

    if return_index:
        return remove_redundant_nodes(lines_out), index_out

    return remove_redundant_nodes(lines_out)


def clip_lines_with_polygon(
    lines: Sequence[LineString],
    polygon: Polygon,
    tolerance: float = 1e-3,
    within: bool = True,
    return_index: bool = False,
) -> Union[Sequence[LineString], Tuple[Sequence[LineString], Sequence[int]]]:
    """clip lines based on polygon outline"""
    # get boundaries of polygon
    boundaries = explode_polygons(polygon)
    # find intersection points of boundaries and lines and split lines based on it
    int_points = intersection_points(lines, boundaries)
    lines_split, index = split_lines(
        lines, int_points, tolerance=tolerance, return_index=True
    )

    # select lines that are contained by polygon
    polygon_buffer = polygon.buffer(
        tolerance
    )  # small buffer to allow for use 'within' function
    if within:
        lines_clip = [line for line in lines_split if line.within(polygon_buffer)]
        index = [
            i for i, line in zip(index, lines_split) if line.within(polygon_buffer)
        ]
    else:
        lines_clip = [line for line in lines_split if not line.within(polygon_buffer)]
        index = [
            i for i, line in zip(index, lines_split) if not line.within(polygon_buffer)
        ]
    if return_index:
        return lines_clip, index

    return lines_clip


def intersection_points(
    lines1, lines2=None, tolerance: float = 0.0, min_spacing: float = 0
):
    """creates list with points of line intersections. if intersection is other type than a point,
    it is broken down to points of its coordinates

    :param tolerance:
    :param min_spacing:
    :param lines1: MultiLineString or list of lines
    :param lines2: MultiLineString or list of lines, if None find intersections amongst lines1
    :return: list with shapely points of intersection
    """
    import rtree

    points = []
    tree_idx_pnt = rtree.index.Index()
    ipnt = 0

    if lines2 is None:
        # build spatial index for lines1
        tree_idx = rtree.index.Index()
        lines_bbox = [l.bounds for l in lines1]
        for i, bbox in enumerate(lines_bbox):
            tree_idx.insert(i, bbox)

    # create multilinestring of close-by lines
    for i1, l1 in enumerate(lines1):
        if lines2 is None:
            # find close-by lines based on bounds with spatial index
            hits = list(tree_idx.intersection(lines_bbox[i1]))
            lines_hit = MultiLineString([lines1[i] for i in hits if i != i1])
        else:
            lines_hit = MultiLineString(lines2)

        if tolerance > 0:
            l1 = extend_line(l1, tolerance)

        x = l1.intersection(lines_hit)
        if not x.is_empty:
            if isinstance(x, Point):
                pnts = [x]

            else:
                if isinstance(x, MultiPoint):
                    pnts = [Point(geom) for geom in x]
                elif isinstance(x, (MultiLineString, MultiPolygon, GeometryCollection)):
                    pnts = [Point(coords) for geom in x for coords in geom.coords]
                elif isinstance(x, (LineString, Polygon)):
                    pnts = [Point(coords) for coords in x.coords]
                else:
                    raise NotImplementedError("intersection yields bad type")

            for pnt in pnts:
                if min_spacing > 0:
                    if ipnt > 0:
                        hits = list(tree_idx_pnt.intersection(pnt.bounds))
                    else:
                        hits = []

                    if len(hits) == 0:  # no pnts within spacing
                        ipnt += 1
                        tree_idx_pnt.insert(ipnt, pnt.buffer(min_spacing).bounds)
                        points.append(pnt)
                else:
                    points.append(pnt)

    return points



import logging
import os
import sys
from pathlib import Path
from typing import Any, Optional

__all__ = ["setup_logger"]

level_map = {
    logging.NOTSET: 0,  # Qgis.MessageLevel.NoLevel
    # 0 When set on a logger, indicates that ancestor loggers are to be consulted to determine the effective level. If that still resolves to NOTSET, then all events are logged. When set on a handler, all events are handled.
    logging.DEBUG: 0,  # Qgis.MessageLevel.Info
    # 10 Detailed information, typically only of interest to a developer trying to diagnose a problem.
    logging.INFO: 0,  # Qgis.MessageLevel.Info  #Qgis.MessageLevel.Success
    # 20 Confirmation that things are working as expected.
    logging.WARNING: 1,  # Qgis.MessageLevel.Warning
    # 30 An indication that something unexpected happened, or that a problem might occur in the near future (e.g. ‘disk space low’). The software is still working as expected.
    logging.ERROR: 2,  # Qgis.MessageLevel.Critical
    # 40 Due to a more serious problem, the software has not been able to perform some function.
    logging.CRITICAL: 2,  # Qgis.MessageLevel.Critical
    # 50 A serious error, indicating that the program itself may be unable to continue running.}
}


class QgsLogHandler(logging.Handler):
    """A logging handler that will log messages to the QGIS logging console."""

    def __init__(
        self, tag_name, iface: Optional[Any] = None, level: int = logging.NOTSET
    ):
        self.tag_name = tag_name
        self.iface = iface
        super().__init__(level=level)

    def emit(self, record: logging.LogRecord) -> None:
        """Try to log the message to QGIS if available, otherwise do nothing.

        :param record: logging message containing whatever info needs to be
                logged.
        :type record: str
        """

        # noinspection PyUnresolvedReferences
        from qgis.core import QgsMessageLog

        push = False

        level = level_map[record.levelno]
        if self.iface:
            push = True

        message = f"{record.name}({record.lineno}): {record.getMessage()}"

        QgsMessageLog.logMessage(
            message=message, tag=self.tag_name, notifyUser=push, level=level
        )

        # optionally, display message on QGIS Message bar (above the map canvas)
        if push and level >= logging.WARNING:
            self.iface.messageBar().pushMessage(
                title=self.tag_name,
                text=message,
                level=level,
                duration=(level + 1) * 3,
            )


def add_logging_handler_once(logger: logging.Logger, handler: logging.Handler) -> bool:
    """A helper to add a handler to a logger, ensuring there are no duplicates.

    :param logger: Logger that should have a handler added.
    :type logger: logging.logger

    :param handler: Handler instance to be added. It will not be added if an
        instance of that Handler subclass already exists.
    :type handler: logging.Handler

    :returns: True if the logging handler was added, otherwise False.
    :rtype: bool
    """
    class_name = handler.__class__.__name__

    for handler_ in logger.handlers:
        if handler_.__class__.__name__ == class_name:
            return False

    logger.addHandler(handler)
    return True


def setup_logger(
    logger_name: str,
    *,
    iface: Optional[Any] = None,
    sentry_url: Optional[str] = None,
    log_file: Optional[Path] = None,
    logger_level: int = logging.INFO,
    default_handler_level=logging.DEBUG,
) -> logging.Logger:
    """Run once when the module is loaded and enable logging.

    :param sentry_url: Mandatory url to sentry api for remote logging.
        Consult your sentry instance for the client instance url.
    :type sentry_url: str

    :param log_file: Optional full path to a file to write logs to.
    :type log_file: str

    Borrowed heavily from this:
    http://docs.python.org/howto/logging-cookbook.html

    Use this to first initialise the logger in your __init__.py::

       import custom_logging
       custom_logging.setup_logger('http://path to sentry')

    You would typically only need to do the above once ever as the
    safe model is initialised early and will set up the logger
    globally so it is available to all packages / subpackages as
    shown below.

    In a module that wants to do logging then use this example as
    a guide to get the initialised logger instance::

       # The LOGGER is initialised in utilities.py by init
       import logging
       LOGGER = logging.getLogger('QGIS')

    Now to log a message do::

       LOGGER.debug('Some debug message')

    .. note:: The file logs are written to the user tmp dir e.g.:
       /tmp/23-08-2012/timlinux/logs/qgis.log

    """
    logger = logging.getLogger(logger_name)
    logger.setLevel(logger_level)

    # create formatter that will be added to the handlers
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    qgis_handler = QgsLogHandler(logger_name, iface=iface, level=default_handler_level)
    # qgis_handler.setLevel(default_handler_level)
    qgis_handler.setFormatter(formatter)
    add_logging_handler_once(logger, qgis_handler)

    if False:
        # create console handler with a higher log level
        console_handler = logging.StreamHandler()
        console_handler.setLevel(default_handler_level)
        console_handler.setFormatter(formatter)
        add_logging_handler_once(logger, console_handler)

    if log_file:
        # create syslog handler which logs even debug messages
        if not isinstance(log_file, Path):
            log_file = Path(log_file)

        assert log_file.is_file()

        file_handler = logging.FileHandler(str(log_file))
        file_handler.setLevel(default_handler_level)
        file_handler.setFormatter(formatter)
        add_logging_handler_once(logger, file_handler)

    if False:
        # Sentry handler - this is optional hence the localised import
        # It will only log if pip install raven. If raven is available
        # logging messages will be sent to the sentry host.
        # We will only log exceptions.
        # Enable the 'plugins/use_sentry' QgsSettings option
        # before this will be enabled.
        # noinspection PyUnresolvedReferences
        from qgis.core import QgsSettings

        settings = QgsSettings()
        app = "some_app"
        flag = settings.value(
            key=f"{app}/sentry-logging", defaultValue=False, type=bool
        )

        if flag:
            third_party_path = os.path.abspath(
                os.path.join(os.path.dirname(__file__), "third_party")
            )
            if third_party_path not in sys.path:
                sys.path.append(third_party_path)
            # noinspection PyUnresolvedReferences
            from raven.handlers.logging import (
                SentryHandler,
            )  # Deprecated for  https://github.com/getsentry/sentry-python

            # noinspection PyUnresolvedReferences
            from raven import Client

            client = Client(sentry_url)
            sentry_handler = SentryHandler(client)
            sentry_handler.setFormatter(formatter)
            sentry_handler.setLevel(default_handler_level)
            if add_logging_handler_once(logger, sentry_handler):
                logger.debug("Sentry logging enabled")
        else:
            logger.debug("Sentry logging disabled")

    return logger


IGNORE = """

from qgis.core import QgsNetworkAccessManager, QgsMessageLog, Qgis, QgsApplication
import os, logging

LOGDIR = os.path.dirname(os.path.realpath(__file__))
LOGFILE = os.path.join(LOGDIR, 'log.txt')

def setupLogger(logfile):
    logger = logging.getLogger('logger')
    if (logger.hasHandlers()):
        logger.handlers.clear()

    formatter = logging.Formatter('%(asctime)s  %(name)s  %(levelname)s: %(message)s')
    logger.setLevel(logging.DEBUG)
    fileHandler = logging.FileHandler(logfile)
    fileHandler.setLevel(logging.DEBUG)
    fileHandler.setFormatter(formatter)
    logger.addHandler(fileHandler)
    return logger

logger = setupLogger(LOGFILE)

# Map received messages to logger
def write_log_message(message, tag, level):
    if level == Qgis.Warning:
        logger.warning(message)
    elif level == Qgis.Critical:
        logger.error(message)
    else:
        logger.info(message)

# Connect the message received signal to the utility function
QgsApplication.messageLog().messageReceived.connect(write_log_message)

"""



#!/usr/bin/env python3
from typing import Sequence, Tuple, Optional

import numpy
from shapely.geometry import Polygon, Point, LineString
from shapely.geometry.base import BaseGeometry
from warg import pairs, Number

__all__ = [
    "project_point_to_object",
    "project_point_to_line_points",
    "project_point_to_line",
    "nearest_geometry",
]


def project_point_to_object(point: Point, geometry: BaseGeometry) -> Point:
    """Find the nearest point in geometry, measured from given point.

    :param point: a shapely Point
    :param geometry: a shapely geometry object (LineString, Polygon)

    :return: a shapely Point that lies on geometry closest to point
    """
    nearest_point = None
    min_dist = float("inf")

    if isinstance(geometry, Polygon):
        for seg_start, seg_end in pairs(list(geometry.exterior.coords)):
            line_start = Point(seg_start)
            line_end = Point(seg_end)

            intersection_point = project_point_to_line_points(
                point, line_start, line_end
            )
            cur_dist = point.distance(intersection_point)

            if cur_dist < min_dist:
                min_dist = cur_dist
                nearest_point = intersection_point

    elif isinstance(geometry, LineString):
        for seg_start, seg_end in pairs(list(geometry.coords)):
            line_start = Point(seg_start)
            line_end = Point(seg_end)

            intersection_point = project_point_to_line_points(
                point, line_start, line_end
            )
            cur_dist = point.distance(intersection_point)

            if cur_dist < min_dist:
                min_dist = cur_dist
                nearest_point = intersection_point
    else:
        raise NotImplementedError(
            "project_point_to_object not implemented for"
            + " geometry type '"
            + geometry.type
            + "'."
        )
    return nearest_point


def project_point_to_line_points(
    point: Point, line_start: Point, line_end: Point, must_be_orthogonal: bool = False
) -> Point:
    """Find the nearest point on a straight line, measured from given point.

    Source: http://gis.stackexchange.com/a/438/19627

    :param must_be_orthogonal:
    :param point: a shapely Point object
    :param line_start: the line starting point as a shapely Point
    :param line_end: the line end point as a shapely Point

    :return: a shapely Point that lies on the straight line closest to point

    """

    line_magnitude = line_start.distance(line_end)

    u = (
        (point.x - line_start.x) * (line_end.x - line_start.x)
        + (point.y - line_start.y) * (line_end.y - line_start.y)
    ) / (line_magnitude**2)

    # closest point does not fall within the line segment,
    # take the shorter distance to an endpoint

    if u < 0.00001 or u > 1:
        ix = point.distance(line_start)
        iy = point.distance(line_end)

        if ix > iy:
            if must_be_orthogonal:
                raise Exception
            return line_end

        else:
            if must_be_orthogonal:
                raise Exception()
            return line_start

    ix = line_start.x + u * (line_end.x - line_start.x)
    iy = line_start.y + u * (line_end.y - line_start.y)

    return Point([ix, iy])


def project_point_to_line(point: Point, line: LineString) -> Point:
    line_coords = line.coords

    # assert line_coords == 2

    return project_point_to_line_points(point, *[Point(*xy) for xy in line_coords])


def line_line_intersection(line: LineString, other: LineString) -> Optional[Point]:
    """

    p = p1_start
    r = (p1_end - p1_start)

    q = p2_start
    s = (p2_end - p2_start)

    t = numpy.cross(q - p, s) / (numpy.cross(r, s))

    # This is the intersection point
    i = p + t * r

    :param line:
    :param other:
    :return:
    """

    import sympy.geometry

    l1 = sympy.geometry.Line(*[sympy.geometry.Point(*xy) for xy in line.coords])

    l2 = sympy.geometry.Line(*[sympy.geometry.Point(*xy) for xy in other.coords])

    l1_l2_intersection = l1.intersection(
        l2
    )  # These are two infinite lines defined by two points on the line

    if len(l1_l2_intersection) == 1:
        if isinstance(l1_l2_intersection[0], sympy.geometry.Line2D):  # Same
            return
        return Point(*l1_l2_intersection[0])


def get_intersection_linear_functions(
    a1: Number, b1: Number, a2: Number, b2: Number
) -> Tuple[Number, Number]:
    A = numpy.array([[-a1, 1], [-a2, 1]])
    b = numpy.array([[b1], [b2]])
    # you have to solve linear System AX = b where X = [x y]'
    return numpy.squeeze(numpy.linalg.pinv(A) @ b)  # x,y


def nearest_geometry(
    geometries: Sequence[BaseGeometry], point: Point
) -> Tuple[BaseGeometry, float, int]:
    """Find the nearest geometry among a list, measured from fixed point.

    :param geometries: a list of shapely geometry objects
    :param point: a shapely Point

    :return:        Tuple (geom, min_dist, min_index) of the geometry with minimum distance        to point, its distance min_dist and the list index of geom, so that        geom = geometries[min_index].
    """
    min_dist, min_index = min(
        (point.distance(geom), k) for (k, geom) in enumerate(geometries)
    )

    return geometries[min_index], min_dist, min_index


if __name__ == "__main__":
    print(
        line_line_intersection(
            LineString([[0, 0], [1, 1]]), LineString([[1, 0], [0, 1]])
        )
    )
    print(
        line_line_intersection(
            LineString([[0, 0], [1, 1]]), LineString([[6, 0], [5, 1]])
        )
    )
    print(
        line_line_intersection(
            LineString([[0, 0], [1, 1]]), LineString([[0, 0], [1, 1]])
        )
    )
    print(
        line_line_intersection(
            LineString([[0, 0], [1, 1]]), LineString([[1, 0], [2, 1]])
        )
    )



def gc_layers() -> None:
    # noinspection PyUnresolvedReferences
    import qgis

    # noinspection PyUnresolvedReferences
    from qgis.core import QgsProject

    registry_layers = QgsProject.instance().mapLayers().keys()
    legend_layers = [
        layer.id() for layer in qgis.utils.iface.legendInterface().layers()
    ]
    layers_to_remove = set(registry_layers) - set(legend_layers)
    QgsProject.instance().removeMapLayers(list(layers_to_remove))



import logging
from typing import Any, Optional, Union

# noinspection PyUnresolvedReferences
from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer, QgsProject

# noinspection PyUnresolvedReferences
from qgis.utils import iface

__all__ = ["duplicate_groups", "select_layer_in_group", "is_group_selected"]

from jord.qgis_utilities.helpers.copying import deepcopy_layer

try:
    from types import EllipsisType
except ImportError:
    EllipsisType = type(...)

logger = logging.getLogger(__name__)


def duplicate_groups(
    group_to_duplicate: Any,
    *,
    group_parent: Optional[Any] = None,
    new_name: Union[str, EllipsisType, None] = None,
) -> QgsLayerTreeGroup:
    if group_to_duplicate is None:
        logger.error("Group was None")
        return

    logger.info(f"Duplicating {group_to_duplicate.name()}")

    if new_name is ... or new_name == "" or new_name is None:
        new_name = f"{group_to_duplicate.name()} (Copy)"

    sub_items = []
    assert (
        group_to_duplicate is not None
    ), f"{group_to_duplicate=} is required to create a duplicate group"

    if group_parent is None:
        group_parent = group_to_duplicate.parent()

    if group_parent is None:
        raise ValueError(f"Group parent was not found for {group_to_duplicate}")

    new_group_parent = group_parent.addGroup(new_name)
    for original_group_child in group_to_duplicate.children():
        if isinstance(original_group_child, QgsLayerTreeGroup):
            new_sub_group, sub_sub_items = duplicate_groups(
                original_group_child, group_parent=new_group_parent, new_name=...
            )
            sub_items.extend([new_sub_group, *sub_sub_items])
        elif isinstance(original_group_child, QgsLayerTreeLayer):
            sub_items.append(
                duplicate_tree_node(new_group_parent, original_group_child)
            )
        else:
            logger.error(f"{original_group_child} no supported in duplication")

    return new_group_parent, sub_items


def duplicate_tree_node(new_group_parent: Any, original_group_child: Any) -> Any:
    original_layer = original_group_child.layer()
    new_layer_copy = deepcopy_layer(original_layer)
    QgsProject.instance().addMapLayer(new_layer_copy, False)

    if False:
        new_layer_tree_node = QgsLayerTreeLayer(
            new_layer_copy
        )  # WORKS BUT MISSING STYLING
        new_group_parent.addChildNode(new_layer_tree_node)
    elif False:
        # Does not WORK EITHER
        new_layer_tree_node = QgsLayerTreeLayer(
            new_layer_copy.id(),
            original_layer.name(),
            new_layer_copy.source(),
            original_layer.providerType(),
        )
    elif False:
        # THIS DOES NOT WORK!!!
        new_layer_tree_node = (
            original_group_child.clone()
        )  # THIS JUST CREATES NEW VIEW of the same data!!!!
        new_layer_tree_node.layer().setDataSource(
            new_layer_copy.source(),
            new_layer_copy.name(),
            new_layer_copy.providerType(),
        )
    elif True:
        new_group_parent.insertLayer(0, new_layer_copy)
    else:
        raise Exception()

    return new_layer_copy


def select_layer_in_group(layer_name: Any, group_name: Any) -> None:
    group = QgsProject.instance().layerTreeRoot().findGroup(group_name)
    if group is not None:
        for child in group.children():
            if child.name() == layer_name:
                iface.setActiveLayer(child.layer())


def is_group_selected(group_name: Any) -> Any:
    group = QgsProject.instance().layerTreeRoot().findGroup(group_name)
    return group in iface.layerTreeView().selectedNodes()



#!/usr/bin/env python3
import datetime
import json
import logging
import time
from typing import Any, Iterable, List, Mapping, Optional, Union

from warg import n_uint_mix_generator_builder, passes_kws_to

from jord.geojson_utilities import GeoJsonGeometryTypesEnum

APPEND_TIMESTAMP = True
SKIP_MEMORY_LAYER_CHECK_AT_CLOSE = True
PIXEL_SIZE = 1
DEFAULT_NUMBER = 0
CONTRAST_ENHANCE = True
DEFAULT_LAYER_NAME = "TemporaryLayer"
DEFAULT_LAYER_CRS = "EPSG:4326"
VERBOSE = False
STRICT = False  # TODO: SET TO TRUE!
USE_TEMP_GROUP = False

__all__ = [
    "add_qgis_single_feature_layer",
    "add_qgis_single_geometry_layers",
    "add_qgis_multi_feature_layer",
    "solve_type_configuration",
    "solve_type",
    "solve_field_uri",
]

logger = logging.getLogger(__name__)

ADD_STRING_LEN = True
NUM_MB16_CHARS = 16777216


def solve_type(d: Any) -> str:
    """
    Does not support size/length yet...

    QGUS Availible Field types:
    ________________           - Provider type ( MemoryLayer) - implemented

    Whole Number (integer) - integer - X
    Decimal Number (real) - double - X
    Text (string) - string - X
    Date - date - X
    Time - time - X
    Date & Time - datetime - X
    Whole Number ( ... llint - 16bit) - int2 - O
    Whole Number (integer - 32bit) - int4 - O
    Whole Number (integer - 64bit) - int8 - O
    Decimal Number (numeric) - numeric - O
    Decimal Number (decimal) - decimal - O
    Decimal Number (real) - real - O
    Decimal Number (double) - double precision - O
    Text, unlimited length (text) - text - X
    Boolean - boolean  - X
    Binary Object (BLOB) - binary - X
    String List - stringlist - X
    Integer List - integerlist - O
    Decimal (double) List - doublelist - O
    Integer (64 bit) List - integer64list - O
    Map - map - O
    Geometry - geometry - O

    :param d:
    :return:
    """
    if not isinstance(d, bool):
        if isinstance(d, int):
            return "integer"

        elif isinstance(d, float):
            return "double"

        elif isinstance(d, bytes):
            return "binary"

        elif isinstance(d, (list, tuple, set)):  # ASSUME IS STRINGS
            return "stringlist"

        elif isinstance(d, datetime.datetime):
            return "datetime"

        elif isinstance(d, datetime.date):
            return "date"

        elif isinstance(d, datetime.time):
            return "time"

        elif isinstance(d, str):
            if False:
                if (
                    ADD_STRING_LEN
                ):  # WARNING! Shapefiles have a limitation of maximum 254 characters per field
                    return f"string({min(max(len(d) * 16, 255), NUM_MB16_CHARS)})"  # 16x buffer for large strings
            else:
                return "text"

    if isinstance(d, bool):
        if False:
            if ADD_STRING_LEN:
                return "string(255)"  # True, False (5)
        else:
            return "boolean"

    if False:
        return "text"

    return "string"


def solve_type_configuration(d: Any) -> Optional[str]:
    if isinstance(d, str):
        a = len(d)
        if True:
            a *= 2

        a = max(a, 255)
        return str(a)

    return None


def add_qgis_single_feature_layer(
    qgis_instance_handle: Any,
    geom: Any,  #: QgsGeometry,
    name: Optional[str] = None,
    crs: Optional[str] = None,
    columns: Optional[Mapping[str, Any]] = None,
    index: bool = False,
    categorise_by_attribute: Optional[str] = None,
    color_generator: callable = n_uint_mix_generator_builder(
        255, 255, 255, mix_min=(222, 222, 222)
    ),
    group: Any = None,
    visible: bool = True,
) -> List:
    """
    An example url is “Point?crs=epsg:4326&field=id:integer&field=name:string(20)&index=yes”

    :param color_generator:
    :param visible:
    :param categorise_by_attribute:
    :param group:
    :param columns: Field=name:type(length,precision) Defines an attribute of the layer. Multiple field
    parameters can be added to the data provider definition. Type is one of “integer”, “double”, “string”.
    :param index: index=yes Specifies that the layer will be constructed with a spatial index
    :param qgis_instance_handle:
    :param geom:
    :type geom: QgsGeometry
    :param name:
    :type name: Optional[str]
    :param crs: Crs=definition Defines the coordinate reference system to use for the layer. Definition is any
    string accepted by QgsCoordinateReferenceSystem.createFromString()
    :return: None
    :rtype: None
    """
    # noinspection PyUnresolvedReferences
    from qgis.core import (
        QgsVectorLayer,
        QgsFeature,
        QgsVectorLayer,
        QgsRasterLayer,
        QgsProject,
    )
    from .categorisation import categorise_layer

    # noinspection PyUnresolvedReferences
    import qgis

    # uri = geom.type()
    # uri = geom.wkbType()
    # uri = geom.wktTypeStr()

    return_collection = []

    geom_json = json.loads(geom.asJson())
    if geom_json is None:
        return return_collection

    geom_type = geom_json["type"]
    uri = geom_type  # TODO: URI MIGHT BE NONE?
    if uri is None:
        raise Exception(f"Could not load {geom.asJson()} as json")

    if name is None:
        name = DEFAULT_LAYER_NAME

    if crs is None:
        crs = DEFAULT_LAYER_CRS

    layer_name = f"{name}"
    if APPEND_TIMESTAMP:
        layer_name += f"_{time.time()}"

    if columns:
        fields = {k: solve_type(v) for k, v in columns}
        field_type_configuration = {
            k: solve_type_configuration(v) for k, v in columns.items()
        }
    else:
        fields = None
        field_type_configuration = None

    if categorise_by_attribute and fields:
        assert (
            categorise_by_attribute in fields
        ), f"{categorise_by_attribute} was not found in {fields}"

    if qgis_instance_handle is None:
        qgis_project = QgsProject.instance()
    elif not isinstance(qgis_instance_handle, QgsProject):
        qgis_project = qgis_instance_handle.qgis_project
    else:
        qgis_project = qgis_instance_handle

    if geom_type == GeoJsonGeometryTypesEnum.geometry_collection.value.__name__:
        for g in geom.asGeometryCollection():  # TODO: Look into recursion?
            uri = json.loads(g.asJson())["type"]
            if uri is None:
                raise Exception(f"Could not load {g.asJson()} as json")
            sub_type = uri  # TODO: URI MIGHT BE NONE?

            uri += "?"

            if crs:
                uri += f"crs={crs}"

            if fields:
                uri = str(uri).rstrip("&")
                for k, v in fields.items():
                    uri += f"&field={k}:{v}"

            uri = str(uri).rstrip("&")
            uri += f'&index={"yes" if index else "no"}'
            uri.replace("?&", "?")

            feat = QgsFeature()
            feat.setGeometry(g)

            if columns:
                feat.initAttributes(len(columns))

                if STRICT:
                    for field_idx, attr in enumerate(columns.values()):
                        feat.setAttribute(field_idx, attr)
                else:
                    feat.setAttributes(
                        list(to_string_if_not_of_exact_type(columns.values()))
                    )

            sub_layer = QgsVectorLayer(uri, f"{layer_name}_{sub_type}", "memory")
            layer_data_provider = sub_layer.dataProvider()
            layer_data_provider.addFeatures([feat])
            layer_data_provider.updateExtents()

            if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
                sub_layer.setCustomProperty("skipMemoryLayersCheck", 1)

            if categorise_by_attribute:
                categorise_layer(
                    sub_layer, categorise_by_attribute, iterable=color_generator
                )

            sub_layer.commitChanges()
            sub_layer.updateFields()
            sub_layer.updateExtents()

            return_collection.append(sub_layer)

            if group:
                qgis_project.addMapLayer(sub_layer, False)
                group.insertLayer(0, sub_layer)
            else:
                qgis_project.addMapLayer(sub_layer)

            layer_tree_handle = qgis_project.layerTreeRoot().findLayer(sub_layer.id())
            if layer_tree_handle:
                layer_tree_handle.setItemVisibilityChecked(visible)
    else:
        uri += "?"

        if crs:
            uri += f"crs={crs}"

        if fields:
            uri = solve_field_uri(field_type_configuration, fields, uri)

        uri = str(uri).rstrip("&")
        uri += f'&index={"yes" if index else "no"}'
        uri.replace("?&", "?")

        feat = QgsFeature()
        feat.setGeometry(geom)

        if columns:
            feat.initAttributes(len(columns))

            if False:
                for field_idx, attr in enumerate(columns.values()):
                    feat.setAttribute(field_idx, attr)
            else:
                feat.setAttributes(
                    list(to_string_if_not_of_exact_type(columns.values()))
                )

        layer = QgsVectorLayer(uri, layer_name, "memory")
        layer_data_provider = (
            layer.dataProvider()
        )  # DEFAULT DATA PROVIDER, TODO: MAYBE CHANGE THIS
        assert feat.isValid(), f"{feat} was invalid"
        res, out_feats = layer_data_provider.addFeatures([feat])

        if not res:
            logger.error(f"{layer_data_provider.lastError()}")

            assert (
                res
            ), f"Failure while adding features {res} {layer_data_provider.lastError()}"

        layer_data_provider.updateExtents()

        if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
            layer.setCustomProperty("skipMemoryLayersCheck", 1)

        if categorise_by_attribute:
            categorise_layer(layer, categorise_by_attribute, iterable=color_generator)

        layer.commitChanges()
        layer.updateFields()
        layer.updateExtents()

        return_collection.append(layer)

        if group:
            qgis_project.addMapLayer(layer, False)
            group.insertLayer(0, layer)
        else:
            qgis_project.addMapLayer(layer)

        layer_tree_handle = qgis_project.layerTreeRoot().findLayer(layer.id())
        if layer_tree_handle:
            layer_tree_handle.setItemVisibilityChecked(visible)

    actions = qgis.utils.iface.layerTreeView().defaultActions()
    actions.showFeatureCount()
    actions.showFeatureCount()  # TODO: Duplicate?

    return return_collection


@passes_kws_to(add_qgis_single_feature_layer)
def add_qgis_single_geometry_layers(
    qgis_instance_handle: Any, geoms: Mapping, **kwargs  # [str,QgsGeometry]
) -> None:
    for name, geom in geoms.items():
        add_qgis_single_feature_layer(qgis_instance_handle, geom, name, **kwargs)


def add_qgis_multi_feature_layer(
    qgis_instance_handle: Any,
    geoms: Iterable,  # [QgsGeometry]
    name: Optional[Iterable[str]] = None,
    crs: Optional[str] = None,
    columns: Optional[
        Union[Mapping[str, Mapping[str, Any]], Iterable[Mapping[str, Any]]]
    ] = None,
    categorise_by_attribute: Optional[str] = None,
    color_generator: Iterable = n_uint_mix_generator_builder(
        255, 255, 255, mix_min=(222, 222, 222)
    ),
    index: bool = False,
    group: Any = None,
    visible: bool = True,
) -> Optional[List]:
    """

        fields  == column definition name, type, length/size
        Multiple field parameters can be added to the data provider definition. type is one of “integer”,
        “double”, “string”.

    An example url is “Point?crs=epsg:4326&field=id:integer&field=name:string(20)&index=yes”

    :param color_generator:
    :param categorise_by_attribute:
    :param visible:
    :param group:
    :param qgis_instance_handle:
    :param geoms:
    :param name:
    :param crs:
    :param columns:
    :param index:
    :return:
    """

    from .categorisation import categorise_layer

    # noinspection PyUnresolvedReferences
    from qgis.core import QgsFeature, QgsVectorLayer, QgsProject, QgsFeatureSink

    # noinspection PyUnresolvedReferences
    import qgis

    # uri = geom.type()
    # uri = geom.wkbType()
    # uri = geom.wktTypeStr()

    return_collection = []

    if name is None:
        name = DEFAULT_LAYER_NAME

    if crs is None:
        crs = DEFAULT_LAYER_CRS

    layer_name = f"{name}"
    if APPEND_TIMESTAMP:
        layer_name += f"_{time.time()}"

    geom_type = None
    uri = None

    sample_row = None
    num_cols = None
    attr_generator = None
    fields = None
    field_type_configuration = None

    if columns and len(columns):  # TODO: FIND MAX LENGTH STR
        if isinstance(columns, Mapping):
            sample_row = next(iter(columns.values()), None)
            # TODO: Might not be ordered correctly
            # if isinstance(next(iter(columns.values())), Mapping):
            #    ...
            attr_generator = iter(columns.values())
        elif isinstance(columns, Iterable):
            sample_row = next(iter(columns), None)
            # TODO: Might not be ordered correctly
            # if isinstance(next(iter(columns.values())), Mapping):
            #    ...
            attr_generator = iter(columns)

        assert isinstance(sample_row, Mapping)

    if sample_row:
        fields = {k: solve_type(v) for k, v in sample_row.items()}
        field_type_configuration = {
            k: solve_type_configuration(v) for k, v in sample_row.items()
        }
        num_cols = len(sample_row)

    if categorise_by_attribute and fields:
        assert (
            categorise_by_attribute in fields
        ), f"{categorise_by_attribute} was not found in {fields}"

    if not geoms:
        # logger.info(f"Found no geometries, {geoms} for {name}")
        return  # No geometry

    features = []
    if not isinstance(geoms, Iterable):
        geoms = [geoms]

    for geom in geoms:
        # geom:QgsGeometry
        geom_json = json.loads(geom.asJson())

        if geom_json is None:
            continue

        geom_type_ = geom_json["type"]

        assert geom_type_ is not None, f"could not read {geom_type_=} as json"

        if geom_type_ is None:
            raise Exception(f"Could not load {geom.asJson()} as json")

        if geom_type is None:
            geom_type = geom_type_

        if uri is None:
            uri = geom_type_  # TODO: URI MIGHT BE NONE?

        assert (
            geom_type == geom_type_
        ), f"{geom_type_} is the not the same geometry type as {geom_type}"

        if geom_type == GeoJsonGeometryTypesEnum.geometry_collection.value.__name__:
            for g in geom.asGeometryCollection():  # TODO: Look into recursion?
                sub_type = json.loads(g.asJson())["type"]
                if sub_type is None:
                    raise Exception(f"Could not load {g.asJson()} as json")

                # logger.info(f'Adding {sub_type=} of {geom=}')

                return_collection.extend(
                    add_qgis_multi_feature_layer(
                        qgis_instance_handle, g, f"{name}_{sub_type}"
                    )
                )
            return return_collection
        else:
            feat = QgsFeature()

            if attr_generator:
                row = next(attr_generator, None)
                if row:
                    feat.initAttributes(num_cols)
                    feat.setAttributes(
                        list(to_string_if_not_of_exact_type(row.values()))
                    )

            feat.setGeometry(geom)

            assert feat.isValid(), f"{feat} was invalid"
            features.append(feat)

    if STRICT:
        assert len(list(geoms)) == len(
            features
        ), f"Some features where dropped! {len(list(geoms))} != {len(features)}"
    else:
        if len(list(geoms)) != len(features):
            logger.error(
                f"Some features where dropped! {len(list(geoms))} != {len(features)}"
            )

    if uri is None:
        if STRICT:
            raise Exception("uri is None")
        else:
            return return_collection

    uri += "?"

    if crs:
        uri += f"crs={crs}"

    if fields:  # field=name:type(length,precision)
        uri = solve_field_uri(field_type_configuration, fields, uri)

    uri = str(uri).rstrip("&")
    uri += f'&index={"yes" if index else "no"}'
    uri.replace("?&", "?")

    layer = QgsVectorLayer(uri, layer_name, "memory")
    layer_data_provider = layer.dataProvider()
    # pr.addAttributes([QgsField("name", QVariant.String),QgsField("age", QVariant.Int),QgsField("size",
    # QVariant.Double)])

    (res, out_feats) = layer_data_provider.addFeatures(
        features
        # , QgsFeatureSink.RollBackOnErrors
    )

    if not res:
        logger.error(f"{layer_data_provider.lastError()}")

        assert (
            res
        ), f"Failure while adding features {res} {layer_data_provider.lastError()}"

        assert len(list(geoms)) == len(
            out_feats
        ), f"Some features where dropped! return status {res}:  {len(list(geoms))} != {len(out_feats)}"

    if len(list(geoms)) != layer.featureCount():
        logger.error(f"{features}")

    if STRICT:
        assert (
            len(list(geoms)) == layer.featureCount()
        ), f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
    else:
        if len(list(geoms)) != layer.featureCount():
            logger.error(
                f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
            )

    layer_data_provider.updateExtents()

    if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
        layer.setCustomProperty("skipMemoryLayersCheck", 1)

    if categorise_by_attribute:
        categorise_layer(layer, categorise_by_attribute, iterable=color_generator)

    if STRICT:
        assert (
            len(list(geoms)) == layer.featureCount()
        ), f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
    else:
        if len(list(geoms)) != layer.featureCount():
            logger.error(
                f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
            )

    layer.commitChanges()
    layer.updateFields()
    layer.updateExtents()

    if STRICT:
        assert (
            len(list(geoms)) == layer.featureCount()
        ), f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
    else:
        if len(list(geoms)) != layer.featureCount():
            logger.error(
                f"Some features where dropped! {len(list(geoms))} != {layer.featureCount()}"
            )

    if qgis_instance_handle is None:
        qgis_project = QgsProject.instance()
    elif not isinstance(qgis_instance_handle, QgsProject):
        qgis_project = qgis_instance_handle.qgis_project
    else:
        qgis_project = qgis_instance_handle

    if group:
        qgis_project.addMapLayer(layer, False)
        group.insertLayer(0, layer)
    else:
        qgis_project.addMapLayer(layer)

    layer_tree_handle = qgis_project.layerTreeRoot().findLayer(layer.id())
    if layer_tree_handle:
        layer_tree_handle.setItemVisibilityChecked(visible)

    actions = qgis.utils.iface.layerTreeView().defaultActions()
    actions.showFeatureCount()  # TODO: Twice?
    actions.showFeatureCount()

    return_collection.append(layer)

    assert len(return_collection) > 0, f"Return collection was empty"

    return return_collection


def solve_field_uri(field_type_configuration, fields, uri):
    uri = str(uri).rstrip("&")
    for k, v in fields.items():
        uri += f"&field={k}:{v}"
        if field_type_configuration is not None and k in field_type_configuration:
            c = field_type_configuration[k]
            if c:
                uri += f"({c})"
    return uri


def to_string_if_not_of_exact_type(
    gen: Iterable, type_: Iterable[type] = (int, float)
) -> Union[str, Any]:
    """

    :param type_: Type for testing against
    :param gen: The iterable to be converted
    :return:
    """
    if not isinstance(type_, Iterable):
        type_ = [type_]

    for v in gen:
        if all([type(v) != t for t in type_]) or v is None:
            yield str(v)
        else:
            yield v



#!/usr/bin/env python3

__author__ = "heider"
__doc__ = r"""

           Created on 5/5/22
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md") as this_init_file:
    __doc__ += this_init_file.read()


from .procedures import *
from .serialisation import *
from .uri_utilities import *
from .client import *
from .clients import *
